/*
 * NBCTK
 * Copyright (c) 2007-2011, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

static void split_name(const char *fname, int line_c, char *buf, char *buf0,
                       char *name, char *values)
{
  char *b,*n,*v;
  int state;

  b = buf;
  n = name;
  v = values;
  state = 0;
  while (*b != '\0') {
    if (state == 0) {
      if (!isspace(*b))
        *(n++) = *b;
      else
        state++;
      b++;
    }
    else if (state == 1) {
      if (!isspace(*b))	state++;
      else b++;
    }
    else
      *(v++) = *(b++);
  }
  *n = *v = '\0';

  if (state != 2)
    die_line(fname,line_c,__LINE__,buf0,"Unexpected format",-1);
}

/* we check if this is an empty line, and skip if it is the case.
 * additionally, the comment part will be deleted and the rest part
 * will be trimmed.
 */
static int check_empty_line(char *buf)
{
  char *b;
  int is_empty;

  b = buf;
  is_empty = 1;
  while (*b != '\0') {
    if (*b == '#' || *b == '\n' || *b == '\r') {
      *b = '\0'; /* delete the comment */
      break;
    }
    else if (!isspace(*b)) {
      is_empty = 0;
      break;
    }
    b++;
  }
  if (!is_empty) trim(buf," \t\r\n",1);

  return is_empty;
}

static enum name_type classify_name(char *values)
{
  if (strcmp(values,"id")      == 0) return ANAME_ID;
  if (strcmp(values,"eval")    == 0) return ANAME_EVAL;
  if (strcmp(values,"ignored") == 0) return ANAME_IGN;
  if (strcmp(values,"integer") == 0) return ANAME_CONT_INT;
  if (strcmp(values,"real")    == 0) return ANAME_CONT_REAL;
  if (strcmp(values,"nominal") == 0) return ANAME_DISC_ASIS;
  if (values[0] == '{' && values[strlen(values) - 1] == '}')
    return ANAME_DISC_GIVEN;

  return -1;
}

static void check_name_type(int v, int t, char *name)
{
  int ok = 1;

  switch (t) {
  case ANAME_ID:
    ok = is_attr_id(v);
    break;
  case ANAME_EVAL:
    ok = is_attr_eval(v);
    break;
  case ANAME_IGN:
    ok = is_attr_ignore(v);
    break;
  case ANAME_CONT_INT:
  case ANAME_CONT_REAL:
    ok = is_attr_cont(v);
    break;
  case ANAME_DISC_ASIS:
  case ANAME_DISC_GIVEN:
    ok = is_attr_disc(v);
    break;
  }

  if (!ok) die("Inconsistent attribute type for `%s'\n",name);
}

static void register_disc_given(const char *fname, int line_c, char *buf,
																char *name, int v, char *values)
{
  char *word,*b,*w;
  int v_d,c;

  v_d = get_local_disc_id(v);

  word = alloc_str(MAX_WORD_LEN);

  b = values + 1;
  w = word;
  c = 0;
  while (*b != '}') {
    if (isspace(*b)) {
      /* do nothing */
    }
    else if (*b == ',') {
      *w = '\0';
			if (strlen(word) == 0)
				die_line(fname,line_c,__LINE__,buf,"Empty value name",-1);
      register_hashtable(&ht_word[v_d],word);
      w = word;
    }
    else {
      if (c == MAX_WORD_LEN)
				die_line(fname,line_c,__LINE__,buf,"Too long value name",-1);
      *(w++) = *b;
      c++;
    }
    b++;
  }
  *w = '\0';
	if (strlen(word) == 0)
		die_line(fname,line_c,__LINE__,buf,"Empty value name",-1);
  register_hashtable(&ht_word[v_d],word);

  xfree(word);
}

/* returns the number of name entries */
static int read_name_body(const char *fname, FILE *fp)
{
  char *buf,*buf0;
  int is_empty,v,line_c;
  char buf_name[MAX_LINE_LEN];
  char buf_vals[MAX_LINE_LEN];
  enum name_type t;

  buf  = alloc_str(MAX_LINE_LEN);
  buf0 = alloc_str(MAX_LINE_LEN);

  attr_name = (char **)xmalloc(sizeof(char *)*tuple_size);
  name_type = (enum name_type *)xmalloc(sizeof(enum name_type)*tuple_size);

  line_c = 0;
  v = 0;
  while (fgets(buf,MAX_LINE_LEN,fp) != NULL) {
    line_c++;

    strcpy(buf0,buf);
    is_empty = check_empty_line(buf);
    if (is_empty) continue;

    split_name(fname, line_c, buf, buf0, buf_name, buf_vals);
    t = classify_name(buf_vals);

    if ((int)t < 0)
      die_line(fname,line_c,__LINE__,buf0,"Unexpected attribute spec",-1);

    if (v == tuple_size)
      die_line(fname,line_c,__LINE__,buf0,"Too many attribute names",-1);

    check_name_type(v,t,buf_name);

    attr_name[v] = dup_str(buf_name);
    name_type[v] = t;

    if (t == ANAME_DISC_GIVEN)
			register_disc_given(fname,line_c,buf,buf_name,v,buf_vals);

    v++;
  }

  if (v != tuple_size)
    die_line(fname,line_c,__LINE__,buf0,"Too few attribute names",-1);

  xfree(buf);
  xfree(buf0);

  return 0;
}

/* return 1 if there exists the naming file, 0 otherwise */
void read_name(void)
{
  FILE *fp_name;

  fp_name = fopen(fname_name, "rb");
  if (fp_name == NULL) return;

  print_msg("Reading names from %s...", fname_name);
  read_name_body(fname_name, fp_name);
  print_msg("done\n");

  mode_name = 1;
}
