/*
 * NBCTK
 * Copyright (c) 2007-2011, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

#ifndef MAX_LINE_LEN
#define MAX_LINE_LEN (1048576)
#endif

char **alloc_field(int num_field)
{
  int f;
  char **field;

  field = (char **)xmalloc(sizeof(char *)*num_field);
  for (f=0; f<num_field; f++)
    field[f] = alloc_str(MAX_LINE_LEN);

  return field;
}

void release_field(char **field, int num_field)
{
  int f;

  for (f=0; f<num_field; f++)
    xfree(field[f]);
  xfree(field);
}

/* the returned string should be freed later */
static char *strcat2(const char *s0, const char *s1)
{
  int l0 = (s0 == NULL) ? 0 : strlen(s0);
  int l1 = (s1 == NULL) ? 0 : strlen(s1);
  char *s = alloc_str(l0 + l1);
  if (s0 != NULL) strcat(s,s0);
  if (s1 != NULL) strcat(s,s1);
  return s;
}

/* the returned string should be freed later (if overwrite == 0) */
char *trim(char *buf, char *skip_chars, int overwrite)
{
  char *trimmed = dup_str(buf);
  int l,i,u,v,k;

  l = strlen(buf);
  for (u=0; u<l; u++)
    if (!strchr(skip_chars,buf[u])) break;
  for (v=(l-1); v>=0; v--)
    if (!strchr(skip_chars,buf[v])) break;

  l = v - u + 1;
  if (l <= 0) return 0;

  k = 0;
  for (i=u; i<=v; i++) {
    trimmed[k] = buf[i];
    k++;
  }
  trimmed[k] = '\0';

  if (overwrite) return strcpy(buf,trimmed);
  return trimmed;
}

/*
 * [Note]
 * - split() returns -1 if the number of fields exceeds max_field,
 *   and returns the number of fields found otherwise.
 * - the position of each field will be recorded into pos[].
 *   On the other hand, if pos == NULL, no positions will be recorded.
 */
int split(int max_field, char *buf, char **field, int *pos,
          const char *delims, const char *skip_chars, int ignore_empty)
{
  int c,n,start;
  char *b,*q,*buf_t,*skip_chars0=NULL;

  for (n=0; n<max_field; n++)
    if (pos != NULL) pos[n] = -1;

  if (ignore_empty) {
    skip_chars0 = strcat2(delims,skip_chars);
    buf_t = trim(buf,skip_chars0,0);
  }
  else {
    buf_t = buf;
  }

  n = c = 0;
  b = buf_t;
  q = field[n];
  start = 1;
  while (c <= MAX_LINE_LEN && *b != '\0') {
    if (skip_chars != NULL && strchr(skip_chars,*b)) {
      /* do nothing */
    }
    else if (delims != NULL && strchr(delims,*b)) {
      if (ignore_empty && q == field[n]) {
        /* do nothing */
      }
      else {
        *q = '\0';
        n++;
        if (n == max_field) {
          if (ignore_empty) xfree(buf_t);
          return -1;
        }
        q = field[n];
      }
      start = 1;
    }
    else if (*b == '\0' || *b == '\r' || *b == '\n') {
      *q = '\0';
      break;
    }
    else {
      *q = *b;  /* copy */
      if (start && pos != NULL) pos[n] = (b - buf);
      start = 0;
      q++;
    }
    b++;
    c++;
  }
  *q = '\0';

  if (ignore_empty) {
    xfree(skip_chars0);
    xfree(buf_t);
  }

  return (n + 1);
}

int split_comma(int max_field, char *buf, char **field, int *pos)
{
  return split(max_field,buf,field,pos,","," \t",0);
}

int split_tab(int max_field, char *buf, char **field, int *pos)
{
  return split(max_field,buf,field,pos,"\t"," ",0);
}
